
/******************************************************************************
Identification des paramètres d'un moteur à courant continu
Ce programme est destiné à fonctionner avec l'ensemble de 
commande de moteur à courant continu, disponible à l'adresse suivante:
http://boutique.3sigma.fr/robots/23-experience-commande-de-moteur-electrique.html

Auteur: 3Sigma
Version 1.0 - 24/01/2017
*******************************************************************************/

// Inclusion d'une bibliothèque permettant de lire et d'écrire plus rapidement sur les entrées-sorties digitales.
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/digitalWriteFast.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <digitalWriteFast.h>

// Inclusion d'une bibliothèque permettant de gérer les interruptions externes
// et le "PinChange Interrupts"
// Télécharger à l'adresse http://www.3sigma.fr/telechargements/EnableInterrupt.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
#include <EnableInterrupt.h>

// Inclusion d'une bibliothèque permettant l'exécution à cadence fixe
// d'une partie du programme. Télécharger à l'adresse http://www.3sigma.fr/telechargements/FlexiTimer2.zip
// et décompresser dans le sous-répertoire « libraries » de votre installation Arduino
// Pour plus de détails, voir les pages (en anglais): 
// http://www.arduino.cc/playground/Main/FlexiTimer2
// https://github.com/wimleers/flexitimer2
#include <FlexiTimer2.h>

// Taille des tableaux de mesure
#define arraysize 150

// Définitions et déclarations pour le codeur incrémental
#define codeurPinA 2
#define codeurPinB 3
volatile long ticksCodeur = 0;

// Définition pour le moteur arrière gauche du robot T-Quad (http://boutique.3sigma.fr/12-robots)
#define directionMoteur  4
#define pwmMoteur  5
#define mesurePWM  8
#define mesureCourant 0

int adc = 512;      // Mesure analogique (le 0 physique correspond à 512)
float courant[arraysize];      // Tableau des courants moyens du moteur
int omega[arraysize];      // Tableau des mesures de vitesse du moteur
unsigned long tval[arraysize];      // Tableau des mesures de temps
unsigned long firstTime;
long nbData = 0;
int over = 0;
int idecim = 0;

float courantsomme = 0.0;
long compteurMesureCourant = 0L;

float Vref = 5.;
float tensionNeutreCapteur = 2.5;
float dt = 0.01;


void setup() {
  Serial.begin(115200);
  
  // Codeur incrémental
  pinMode(codeurPinA, INPUT_PULLUP);      // entrée digitale pin A codeur
  pinMode(codeurPinB, INPUT_PULLUP);      // entrée digitale pin B codeur

  // Moteur à courant continu arrière gauche
  pinMode(directionMoteur, OUTPUT);
  pinMode(pwmMoteur, OUTPUT);
  pinMode(mesurePWM, INPUT);

  // Modification du prescaler de l'ADC: division par 8 de la fréquence du micro-contrôleur
  // au lieu de 128 par défaut, pour diminuer le temps entre deux mesures analogiques
  bitClear(ADCSRA, ADPS2);
  bitSet(ADCSRA, ADPS1);
  bitSet(ADCSRA, ADPS0);

  // Calcul du "zéro" du capteur (normalement, 2.5V correspond à un courant nul
  // mais il est préférable de le recalculer dans le programme au cas où il y aurait un offset)
  adc = analogRead(mesureCourant);
  tensionNeutreCapteur = 0.0;
  for (int i=0; i<10; i++) {
    adc = analogRead(mesureCourant);
    tensionNeutreCapteur += adc;
  }
  tensionNeutreCapteur = Vref * (tensionNeutreCapteur / 10.0) / 1024.;
  Serial.print("Tension neutre: ");
  Serial.println(tensionNeutreCapteur);

  FlexiTimer2::set(1, 1/10000., isrt); // résolution timer = 0.1 ms
  FlexiTimer2::start();

  // On ne génère pas un PWM car la période est trop petite
  digitalWrite(pwmMoteur, HIGH);
}

void loop() {
  // Première phase: mesure du courant alors que le moteur
  // ne tourne pas encore
  if ((nbData > (arraysize/2)-1) && (over == 0)) {
    over = 1;
    
    // Arrêt des moteurs
    digitalWrite(pwmMoteur, LOW);
  }
  // Fin de la première phase: transmission des mesures de courant
  // via la liaison série
  else if (over == 1) {
    over = 2;

    // On met arbitrairement le premier et le dernier point à 0
    // pour que l'identification puisse se faire correctement dans MyViz
    courant[0] = 0.0;
    courant[nbData - 1] = 0.0;
    firstTime = tval[0];
    for (int i=0; i<arraysize/2; i++) {
      Serial.print(((float)(tval[i] - firstTime))/1000,3);
      Serial.print(",");
      Serial.print(courant[i], 3);
      Serial.print(",");
      Serial.println((float)omega[i], 3);
    }
    
    // A chaque changement de niveau de tension sur le pin A du codeur,
    // on exécute la fonction GestionInterruptionCodeurPinA (définie à la fin du programme)
    enableInterrupt(codeurPinA, GestionInterruptionCodeurPinA, CHANGE);
    // A chaque changement de niveau de tension sur le pin B du codeur,
    // on exécute la fonction GestionInterruptionCodeurPinB (définie à la fin du programme)
    enableInterrupt(codeurPinB, GestionInterruptionCodeurPinB, CHANGE);
  
    // Calcul courant moteur, sur interruption PWM
    enableInterrupt(mesurePWM, calculCourantMoteur, RISING);

    // Reinitialisations pour la phase suivante
    delay(100);
    ticksCodeur = 0;
    over = 3;

    // Consigne de vitesse moteur (correspond à environ 6V)
    analogWrite(pwmMoteur, 185);    
  }
  // Fin de la seconde phase: transmission des vitesses et mesures de courant
  // via la liaison série
  else if ((over == 3) && (nbData > arraysize-2)) {
    over = 4;
    // Arrêt du moteur
    analogWrite(pwmMoteur, 0);
    
    // On met arbitrairement les dernier points à 0 pour que l'identification
    // puisse se faire correctement dans MyViz
    courant[nbData - 1] = 0.0;
    omega[nbData - 1] = 0;
    for (int i=(arraysize/2); i<arraysize-1; i++) {
      Serial.print(((float)(tval[i] - firstTime))/1000,3);
      Serial.print(",");
      Serial.print(courant[i],3);
      Serial.print(",");
      // Calcul de la vitesse de rotation. C'est le nombre d'impulsions converti en radian, divisé par la période d'échantillonnage
      Serial.println(((2.*3.141592*((float)omega[i]))/1632.)/dt,2);  
    }
    
    // Arrêt de l'interruption Timer 2
    FlexiTimer2::stop();

  }
}

// Routine exécutéee sur interruption Timer 2
void isrt() {
  // Première phase
  if (over == 0) {
    // Mesure du courant et stockage dans un tableau
    adc = analogRead(mesureCourant);
    tval[nbData] = micros();
    courant[nbData] = (Vref * ((float)adc)/1024. - tensionNeutreCapteur) / 0.185;
    nbData++;
  }
  // Seconde phase
  else if (over == 3) {
    adc = analogRead(mesureCourant);  
    courantsomme += (Vref * (float)adc/1024. - tensionNeutreCapteur) / 0.185;
    compteurMesureCourant++;
    // Pour la vitesse, une exécution sur 100 interruptions Timer 2 (échantillonage à 10 ms au lieu de 100 µs)
    if (idecim == 100) {
      idecim = 0;
      tval[nbData] = micros();
      
      // Nombre de ticks codeur depuis la dernière fois: correspond à l'angle de rotation du moteur pendant la période d'échantillonnage
      // Ce nombre est mis à jour par les fonctions GestionInterruptionCodeurPinA et GestionInterruptionCodeurPinB,
      // exécutées à chaque interruption due à une impulsion sur la voie A ou B du codeur incrémental
      omega[nbData] = ticksCodeur;
  
      // Une fois lu, ce nombre est remis à 0 pour pouvoir l'incrémenter de nouveau sans risquer de débordement de variable
      ticksCodeur = 0;
    
      nbData++;
    }
    else {
      idecim++;
    }
  }


}

void calculCourantMoteur() {
  // Calcul du courant moyen sur la période PWM haut / PWM bas écoulée.
  if (compteurMesureCourant > 0) {
    courant[nbData] = courantsomme / (float)compteurMesureCourant;
    compteurMesureCourant = 0L;
  }
  else {
    courant[nbData] = 0.;
  }
  courantsomme = 0.0;
}

void GestionInterruptionCodeurPinA() {  
  // Routine de service d'interruption attachée à la voie A du codeur incrémental
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurPinA) == digitalReadFast2(codeurPinB)) {
    ticksCodeur++;
  }
  else {
    ticksCodeur--;
  }
}

void GestionInterruptionCodeurPinB() {
  // Routine de service d'interruption attachée à la voie B du codeur incrémental
  // On utilise la fonction digitalReadFast2 de la librairie digitalWriteFast
  // car les lectures doivent être très rapides pour passer le moins de temps possible
  // dans cette fonction (appelée de nombreuses fois, à chaque impulsion de codeur)
  if (digitalReadFast2(codeurPinA) == digitalReadFast2(codeurPinB)) {
    ticksCodeur--;
  }
  else {
    ticksCodeur++;
  }
}

